static char *cvs_id =" $Id: xnd_sets.c,v 1.2 2004/07/30 08:04:48 berar Exp $ ";
/* xnd_sets.c    rules for selecting lines sets  */
/* --------------------------------------------- */
/* this translates and calculates small rules    */
/* according to                                  */
/* logical  l :                                  */
/*    (l), l && l, l || l, !l, l_prim            */
/* l_prim :                                      */
/*    true, false, n == n, n != n, n > n,        */
/*    n >= n, n < n, n <= n                      */
/* numerical n:                                  */
/*    (n), n + n, n - n, n * n, n / n            */
/*     n % n, +n, -n, n_prim                     */
/* n_prim :                                      */
/*    H, K, L, integer                           */
/*                                               */
/* --------------------------------------------- */
#define  USE_BRAGG
#define  USE_KEYS
#define  USE_SETS
#define  USE_SETS_RULE 
#include "../xnd_def.h" 
#include <string.h>
#include <ctype.h>
/* --------------------------------------------- */
TYPE_SETS * current_sets;
/* --------------------------------------------- */
static int    syntax_error = 0;
static int    current_error;
static char * current_string;
/* --------------------------------------------- */
static int rule_evaluate( SETS_RULE_PTR rule);
static TYPE_SETS_RULE * rule_analyse( char * string);
static int sets_analyse_error( char * m, char * l);
static int is_enclosed (char * string);
static int get_prim (char * string, TYPE_SETS_RULE *rule);
static int get_oper1 (char * string, TYPE_SETS_RULE *rule, TYPE_SETS_OP op_type, char *op_string);
static int get_oper2 (char * string, TYPE_SETS_RULE *rule, TYPE_SETS_OP op_type, char *op_string);
/* --------------------------------------------- */
int sets_analyse (TYPE_SETS *sets)
{  if (sets == NULL) return 0;
   push_err_msg("sets_analyse (xnd_sets.c)");
   fprintf(list_fich," added rules : ");
	if ( sets->allow_str && sets->allow_str[0] )
   {  fprintf(list_fich," ALLOW='%s' ", sets->allow_str);
	   current_string = sets->allow_str;
      current_error = 0;
      sets->allow =  rule_analyse( sets->allow_str);
      } else sets->allow = NULL;
   if ( sets->deny_str && sets->deny_str[0] ) 
   {  fprintf(list_fich," DENY='%s'", sets->deny_str);
	   current_string = sets->deny_str;
      current_error = 0;
      sets->deny  =  rule_analyse( sets->deny_str);
		} else sets->deny = NULL;
   if ( sets->deny== NULL && sets->allow==NULL ) fprintf(list_fich," NONE");
   fprintf(list_fich,"\n");
	pop_err_msg();
	return syntax_error;
   }
/* --------------------------------------------- */
/* allowed hkl lines return 0, others 1; used to skip lines   */
int sets_extinction(TYPE_SETS *sets)
{  push_err_msg("sets_extinction (xnd_sets.c)");
   if (sets->allow  && rule_evaluate(sets->allow) == 0  ) return 1;
   if (sets->deny && rule_evaluate(sets->deny) != 0 )     return 1;
	pop_err_msg();
   return 0;
   }
/* --------------------------------------------- */
TYPE_SETS * sets_make(TYPE_SETS * previous, int total, int nature)
{  int i;
	TYPE_SETS *sets, *l_sets;
	push_err_msg("sets_make (xnd_sets.c)");
	if (previous)
   {   sets = (TYPE_SETS * ) realloc(previous, total*sizeof(TYPE_SETS));
       tst_abort((sets==NULL),"error realloc set");
       }
   else sets = (TYPE_SETS * ) tst_alloc(total, sizeof(TYPE_SETS),"sets");
	l_sets = sets;
	for (i=0; i < total; i++, l_sets++)
	{  memset(l_sets, 0, sizeof(TYPE_SETS));
	   l_sets->nb_hkl = nature;
      }
   pop_err_msg();
	return sets;
   }
/* --------------------------------------------- */
TYPE_SETS * sets_fget(TYPE_SETS * previous, int total, int *result)
{  char key_name[10], deny_str[NAME_LG], allow_str[NAME_LG], loop_str[NAME_LG];
   int i;
	TYPE_SETS *sets, *l_sets;
	push_err_msg("sets_fget (xnd_sets.c)");
	sets = sets_make(previous, total, 0);
   l_sets = sets;
   *result = 0;
	for (i=0; i<total; i++, l_sets++)
	{  sprintf(key_name,"SET_%d", i);
      l_sets->remark = (sint) read_remark(NULL);
		if (get_keys( key_name, NULL,
        "LINES", KEYTYPE_INT, &l_sets->nb_hkl, 0,
        "DENY", KEYTYPE_STRING, deny_str, "",
        "ALLOW", KEYTYPE_STRING, allow_str, "",
        "LOOP", KEYTYPE_STRING, loop_str, "",
        "AUTO", KEYTYPE_INT, &l_sets->auto_intensity, 0,
  			NULL)!= -1) (*result)++; 
	   l_sets->loop_str  = str_alloc(loop_str); 
	   l_sets->deny_str  = str_alloc(deny_str);
	   l_sets->allow_str = str_alloc(allow_str);
	   }
	fprintf(list_fich,"\n");
	pop_err_msg();
	return sets;
	}
/* --------------------------------------------- */
void sets_new (TYPE_SETS *sets, int number)
{  char key_name[10];
   int i;
	TYPE_SETS * l_sets;
	push_err_msg("sets_new (xnd_sets.c)");
	for (i=0, l_sets = sets; i < number; i++, l_sets++)
	{  if(l_sets->remark) write_remark(l_sets->remark);
      sprintf(key_name,"SET_%d", i);
      if ( l_sets->nb_hkl<0) l_sets->nb_hkl = 0;
      if (new_keys(key_name, 
		    "LINES",  KEYTYPE_INT, EXT_KEY, l_sets->nb_hkl, 0,
          "DENY",  KEYTYPE_STRING, EXT_KEY, l_sets->deny_str, (char *) "",
          "ALLOW", KEYTYPE_STRING, EXT_KEY, l_sets->allow_str, (char *) "",
          "LOOP",  KEYTYPE_STRING, EXT_KEY, l_sets->loop_str, (char *) "",
          "AUTO",  KEYTYPE_INT,   EXT_KEY,  l_sets->auto_intensity, 0,
			NULL) == -1)
			fprintf(stderr, "key @SET_d is necessary to write sets rule\n");
	   }
	pop_err_msg();
	}
/* ------------------------------------------------------------------ */
/*  following are static internal procedures     */
/* --------------------------------------------- */
static int rule_evaluate( SETS_RULE_PTR rule)
{  int eval;
	switch (rule->op)
   {  case OPER_L_NOT : 
	       eval = ! rule_evaluate(rule->left.ptr);  break;
      case OPER_L_AND : 
		    eval = rule_evaluate(rule->left.ptr) && rule_evaluate(rule->right); break;
      case OPER_L_OR  : 
		    eval = rule_evaluate(rule->left.ptr) || rule_evaluate(rule->right); break;
      case OPER_N_EQ  : 
		    eval = rule_evaluate(rule->left.ptr) == rule_evaluate(rule->right); break;
      case OPER_N_NE  : 
		    eval = rule_evaluate(rule->left.ptr) != rule_evaluate(rule->right); break;
      case OPER_N_GT  : 
		    eval = rule_evaluate(rule->left.ptr) >  rule_evaluate(rule->right); break;
      case OPER_N_GE  : 
		    eval = rule_evaluate(rule->left.ptr) >= rule_evaluate(rule->right); break;
      case OPER_N_LT  : 
		    eval = rule_evaluate(rule->left.ptr) <  rule_evaluate(rule->right); break;
      case OPER_N_NEG : 
		    eval = - rule_evaluate(rule->left.ptr);   break;
      case OPER_N_ADD : 
		    eval = rule_evaluate(rule->left.ptr) + rule_evaluate(rule->right); break;
      case OPER_N_SUB : 
		    eval = rule_evaluate(rule->left.ptr) - rule_evaluate(rule->right); break;
      case OPER_N_MUL : 
		    eval = rule_evaluate(rule->left.ptr) * rule_evaluate(rule->right); break;
      case OPER_N_DIV : 
		    eval = rule_evaluate(rule->left.ptr) / rule_evaluate(rule->right); break;
      case OPER_N_MOD : 
		    eval = rule_evaluate(rule->left.ptr) % rule_evaluate(rule->right); break;
      case OPER_P_HKL : 
		   eval = act->h[rule->left.val] ;  break;
      case OPER_P_NUM : 
		   eval =  rule->left.val ;         break;
      default : ;
      }
   return eval;
   }
/* --------------------------------------------- */
static int sets_analyse_error( char * m, char * l)
{  verbose_printf( "syntax error in '%s' : %s '%s'\n", current_string, m, l);
   syntax_error++;
   current_error++;
   return 0;
   }
/* --------------------------------------------- */
/* --------------------------------------------- */
static int is_enclosed (char * string)
{  char *a = string;
   int level = 0, n = 0;
   if (current_error) return 0;
   while (*a)
   {  switch (*a)
      {  case '(' : level++; break;
         case ')' : level--; break;
         }
      if (level == 0 ) n++;
      a++;
      }
   if (level)  return sets_analyse_error( "unbalanced parenthesis in", string);
   if ( string[0] != '(' || n > 1) return 0;
   a--;
   *a=0;
   return 1;  
   }
/* --------------------------------------------- */
static int get_prim (char * string, TYPE_SETS_RULE *rule)
{  char *a = string;
   int  n;
   if (current_error) return -1;
   while (*a == ' ' ) a++ ;
   if ( *a == 'H' )
   {   rule->op = OPER_P_HKL;
       rule->left.val = 0;
       a++;
       }
   else if ( *a >= 'K' && *a <= 'N' )
   {   rule->op = OPER_P_HKL;
       rule->left.val = *a - 'K'+ 1;
       a++;
       }
   else if ( sscanf(a, "%d", &n) == 1)
   {   rule->op = OPER_P_NUM;
       rule->left.val = n;
       while (isdigit(*a)) a++;
       }
   else return sets_analyse_error("unkown primitive ", a);
   while ( *a == ' ' ) a++ ;
   if ( *a ) return sets_analyse_error("unexpected char ", a);
   return 1;
   }
/* --------------------------------------------- */
static int get_oper1 (char * string, TYPE_SETS_RULE *rule,
   TYPE_SETS_OP op_type, char *op_string)
{  char *a = string;
   int level = 0;
   if (current_error) return -1;
   while (*a == ' ' ) a++ ;
   if ( (strstr(a, op_string) == a) && level == 0 )
   {  a += strlen(op_string);
      if (strlen(a) == 0) return sets_analyse_error("empty right operand of", op_string);
      rule->op = op_type;
      rule->left.ptr  = rule_analyse (a);
      return 1;
      }
   return 0;
   }
/* --------------------------------------------- */
static int get_oper2 (char * string, TYPE_SETS_RULE *rule,
   TYPE_SETS_OP op_type, char *op_string)
{  char *right, *a = string+1;
   int level = 0;
   if (current_error) return -1;
   while (*a)
   {  if (*a == op_string[0] )
      {   if ( (strstr(a, op_string) == a) && level == 0 )
          {  *a = 0;
             right = a + strlen(op_string);
             if (strlen(string) == 0 ) return sets_analyse_error("empty left operand of", op_string);
             if (strlen(right) == 0) return sets_analyse_error("empty right operand of", op_string);
             rule->op = op_type;
             rule->left.ptr  = rule_analyse (string);
             rule->right = rule_analyse (right);
             return 1;
             }
          }
      else switch (*a)
      {  case '(' : level++; break;
         case ')' : level--; break;
         }
      a++;
      }
   if (level)  return sets_analyse_error( "unbalanced parenthesis", string);
   return 0;
   }
/* --------------------------------------------- */
static TYPE_SETS_RULE * rule_analyse( char * string )
{  TYPE_SETS_RULE * rule;
   char *copy, *a;
   rule = (TYPE_SETS_RULE *) malloc(sizeof(TYPE_SETS_RULE));
   rule->op = OPER_NONE;
   a = copy = malloc(1+strlen(string));
   strcpy (copy, string);
   while ( *a && (is_enclosed(a) || *a == ' ')) a++;
   if ( ! get_oper2( a, rule, OPER_L_OR,  "||") )
   if ( ! get_oper2( a, rule, OPER_L_OR,  "|" ) )
   if ( ! get_oper2( a, rule, OPER_L_AND, "&&") )
   if ( ! get_oper2( a, rule, OPER_L_AND, "&" ) )

   if ( ! get_oper2( a, rule, OPER_N_GE,  ">=") )
   if ( ! get_oper2( a, rule, OPER_N_LE,  "<=") )
   if ( ! get_oper2( a, rule, OPER_N_NE,  "!=") )
   if ( ! get_oper2( a, rule, OPER_N_EQ,  "==") )
   if ( ! get_oper2( a, rule, OPER_N_EQ,  "=" ) )
   if ( ! get_oper2( a, rule, OPER_N_GT,  ">") )
   if ( ! get_oper2( a, rule, OPER_N_LT,  "<") )

   if ( ! get_oper1( a, rule, OPER_L_NOT, "!" ) )

   if ( ! get_oper2( a, rule, OPER_N_ADD, "+" ) )
   if ( ! get_oper2( a, rule, OPER_N_SUB, "-" ) )
   if ( ! get_oper2( a, rule, OPER_N_MUL, "*" ) )
   if ( ! get_oper2( a, rule, OPER_N_DIV, "/" ) )
   if ( ! get_oper2( a, rule, OPER_N_MOD, "%" ) )

   if ( ! get_oper1( a, rule, OPER_N_NEG, "-" ) )
   if ( ! get_prim ( a, rule ) )
     sets_analyse_error("unresolved sentence ", a);
   free(copy);
   return rule;
   }
/* --------------------------------------------- */
/* --------------------------------------------- */
