static char *cvs_id =" $Id: xnd_sig.c,v 1.2 2004/07/30 08:04:49 berar Exp $ ";
/* XND_SIG.C */

#include "../xnd_def.h"
#ifdef __TURBOC__
#include <alloc.h>
#endif
#include <string.h>

#include <signal.h>
/* ================================================================ */
/* small stack for messages used in error                           */
static char ** err_msg     = NULL;
static int     err_msg_nb  = 0;
static int     err_msg_pos = 0;
/* ---------------------------------------------------------------- */
void clear_err_msg (P_(void))
{   err_msg_pos = 0;
    }
/* ---------------------------------------------------------------- */
int push_err_msg (P_(char *) msg)
#ifdef NO_PROTO
char *msg;
#endif
{   if (err_msg && err_msg_pos >= err_msg_nb ) 
    {    err_msg_nb +=20;
         err_msg = (char **) realloc(err_msg, err_msg_nb*sizeof(char *));
         }
    else if (err_msg==NULL) 
    {    err_msg_nb +=20;
         err_msg = (char **) malloc(err_msg_nb*sizeof(char *));
         }
    err_msg[err_msg_pos++] = msg;
    return err_msg_pos;
	 }
/* ---------------------------------------------------------------- */
void pop_err_msg (P_(void))
{   if (err_msg_pos > 0 ) err_msg_pos--; 
    }
/* ---------------------------------------------------------------- */
void set_err_msg (P_(int) n)
#ifdef NO_PROTO
int n;
#endif
{   if (n >= 0 && n < err_msg_pos) err_msg_pos = n; 
    }
/* ---------------------------------------------------------------- */
void dump_err_msg (P_(void))
{   int i = err_msg_pos;
    verbose_printf ("\n\n  error_message_stack : \n");
    while ( --i >= 0 )
       verbose_printf (" %2d/ %s\n", i, err_msg[i]);
    fflush (list_fich);
    }
/* ================================================================ */
/*  unix signal are used for trapping errors */
/*  but their names depend on the machine    */
/*  hp-ux implementation  */
/*  notice that SIGFPE is no more generated since posix */
#if ( !defined(SIGBUS) && defined(_SIGBUS) )
#define SIGBUS _SIGBUS
#endif
/* ---------------------------------------------------------------- */
static void error (P_(int) sig,  P_(int) cod)
#ifdef NO_PROTO
int sig, cod;
#endif
{   char message[100];
    char *name = "???";
    char *msg="";
    switch (sig)
    {   
#ifdef SIGINT
        case SIGINT  : name = "SIGINT";   
                       msg = "Interrupt by user";
                       break;
#endif
#ifdef SIGILL
        case SIGILL  : name = "SIGILL";    
                       msg = "Illegal instruction, please report";
                       break;
#endif
#ifdef SIGABRT
        case SIGABRT : name = "SIGABRT";
                       msg = "Software generated abort";
                       break;
#endif
#ifdef SIGFPE
        case SIGFPE  : name = "SIGFPE";   
                       msg = "Floating point exception";
                       break;
#endif
#ifdef SIGBUS
        case SIGBUS  : name = "SIGBUS";   
                       msg = "Bus error, please report";
                       break;
#endif
#ifdef SIGSEGV
        case SIGSEGV  : name = "SIGSEGV";    
                       msg = "Memory segmentation violation, please report";
                       break;
#endif
        }
    verbose_printf("\n Abnormal end : signal %d (%s) %s\n", sig, name, msg);
    fflush (list_fich);
    
    if ( sig == SIGFPE )
    {   name = "???";
	switch (cod)
        {
#ifdef __TURBOC__
		 case 126   : name = "FPE_INTOVFLOW"; break;
		 case 127   : name = "FPE_INTDIV"; break;
		 case 129   : name = "FPE_INVALID"; break;
		 case 131   : name = "FPE_ZERODIVID"; break;
		 case 132   : name = "FPE_OVERFLOW"; break;
		 case 133   : name = "FPE_UNDERFLOW"; break; 
		 case 134   : name = "FPE_INEXACT"; break;
		 case 135   : name = "FPE_STACKFAULT"; break;
		 case 140   : name = "FPE_EXPLICITGEN"; break;
#else
/* definition sun/unix   */
#ifdef FPE_INTDIV_TRAP
		 case FPE_INTDIV_TRAP  : name = "FPE_INTDIV_TRAP"; break;
#endif
#ifdef FPE_CHKINST_TRAP
		 case FPE_CHKINST_TRAP : name = "FPE_CHKINST_TRAP"; break;
#endif
#ifdef  FPE_TRAPV_TRAP 
		 case FPE_TRAPV_TRAP   : name = "FPE_TRAPV_TRAP"; break;
#endif
#ifdef FPEfLTBSUN_TRAP
		 case FPE_FLTBSUN_TRAP : name = "FPE_FLTBSUN_TRAP"; break;
#endif
#ifdef FPEfLTINEX_TRAP
		 case FPE_FLTINEX_TRAP : name = "FPE_FLT_INEX_TRAP"; break;
#endif
#ifdef FPEfLTDIV_TRAP
		 case FPE_FLTDIV_TRAP  : name = "FPE_FLT_DIV_TRAP"; break;
#endif
#ifdef FPEfLTUND_TRAP
		 case FPE_FLTUND_TRAP  : name = "FPE_FLTUND_TRAP"; break;
#endif
#ifdef FPEfLTOPERR_TRAP
		 case FPE_FLTOPERR_TRAP: name = "FPE_FLTOPERR_TRAP"; break;
#endif
#ifdef FPEfLTOVF_TRAP
		 case FPE_FLTOVF_TRAP  : name = "FPE_FLTOVF_TRAP"; break;
#endif
#ifdef FPEfLTNAN_TRAP
		 case FPE_FLTNAN_TRAP  : name = "FPE_FLTNAN_TRAP"; break;
#endif

#endif
	   }  /* switch */
	 verbose_printf (" Floating point error code %d (%s)\n", cod, name);
	 }
     else if ( cod ) verbose_printf (" Signal received with code %d)\n", cod);
    fflush (list_fich);
    dump_err_msg();
    termine(13,"signal");
    }
/* ================================================================ */
/* ================================================================ */
void init_error (P_(void))
{   void ( * handler )();
    handler = error;
#ifdef SIGINT
/*    puts("SIGINT defined"); */
    signal (SIGINT, handler);
#endif
#ifdef SIGILL
/*    puts("SIGILL defined"); */
    signal (SIGILL, handler);
#endif
#ifdef SIGABRT
/*    puts("SIGABRT defined"); */
    signal (SIGABRT, handler);
#endif
#ifdef SIGFPE
/*     puts("SIGFPE defined"); */
   signal (SIGFPE, handler);
#endif
#ifdef SIGBUS
/*    puts("SIGBUS defined"); */
    signal (SIGBUS, handler);
#endif
#ifdef SIGSEGV
/*    puts("SIGSEGV defined"); */
    signal (SIGSEGV, handler);
#endif
    }
/* ================================================================ */
/* ================================================================ */
